#!/usr/bin/env python
#
# Data Quality Checker also known as DQC
#
# Read the README.md for more detailed information
#
#       Barcelona Supercomputing Center - Centro Nacional de Supercomputacion
#       Earth Sciences Department - Computational Earth Sciences
#
#     Authors:
#       Alejandro Alfonso Hernández (alejandro.alfonso@bsc.es) 

import os, sys
import json
import cdsapi
from pathlib import Path
from datetime import datetime
import csv

import random

# Custom packages
from dqc_downloader.utils_download import *

# global vars
today_date = datetime.now()
today_year = str(today_date.year)


"""
It sets some desire parameters to do a simple request to cds api
Note: land_sea_mask variables are invariable over time for
a same model (originating_centre + system), so the requests have been simplified
"""

def simplify_req(c, prod_type=None):

    # Get minimal mask (in size)
    if 'leadtime_hour' in c:
        if  isinstance(c['leadtime_hour'], list):
            c['leadtime_hour'] = c['leadtime_hour'][-1]
    if 'time' in c:
        if  isinstance(c['time'], list):
            c['time'] = c['time'][-1]
    if 'day' in c:
        if  isinstance(c['day'], list):
            c['day'] = c['day'][-1]
    if 'month' in c:
        if  isinstance(c['month'], list):
            last_month = sorted(c['month'], key=lambda x: int(x))[-1]
            c['month'] = last_month
    if 'year' in c:
        if today_year in c['year']:
            c['year'] = today_year
        else:
            c['year'] = c['year'][-1]
    if 'product_type' in c:
        c['product_type'] = prod_type

    return c

"""
It downloads variable land_sea_mask of all datasets for CURRENT year
and last months available
"""
def main():
    
    # Params
    if len(sys.argv) != 3:
        print('USAGE: python download_masks_current_year.py <credentials_file> <csv_log_file>')
        exit()
    else:
        cred_path = sys.argv[1]
        csv_path = sys.argv[2]
    
    # URL and key to download
    with open(cred_path) as json_file:
        cred = json.load(json_file)

    if not os.path.exists(os.path.dirname(csv_path)):
        os.makedirs(os.path.dirname(csv_path))
    
    # General configuration
    download_options = 'conf_monthly_updates/download_options_monthly_updates.json'
    download_options = os.path.join(os.path.dirname(os.path.abspath(__file__)), download_options)
    with open(download_options) as json_file:
        conf = json.load(json_file)
        masks_dir = conf['masks_dir']
        logdir = conf['logdir']
        datasets_info_dir = conf['datasets_info_dir']
        cds_forms_c3s_location = Path(conf['formsdir'])
        
    if not(os.path.exists(masks_dir)):
        os.makedirs(masks_dir)
    
    csv_data = defaultdict(list) # dicc that will contain data to be ingest to csv
    csv_date = today_date.strftime("%Y-%m-%d %H:%M:%S")
    
    # Gather masks
    mask_var = 'land_sea_mask'
    downloads = []
    
    for f in os.listdir(datasets_info_dir):
        if f.endswith('.json') and not f.startswith('download'):
            json_path = os.path.join(datasets_info_dir, f)
            # Read list of products to download
            with open(json_path) as json_file:
                data = json.load(json_file)

            for ds in data['dataset']:

                constraints = get_constraints(ds, cds_forms_c3s_location, "constraints")
                for c in constraints:
                    
                    if mask_var in c['variable'] and today_year in c['year']:

                        c['variable'] = mask_var
                        condition = False
                        if 'seasonal' in ds:
                            origin = data['form']['originating_centre']
                            if origin in c['originating_centre']:
                                last_system = sorted(c['system'], key=lambda x: int(x))[-1]
                                c['system'] = last_system
                                c['originating_centre'] = origin
                                condition = True

                            if condition:
                                outfile = '{dataset}-{origin}-{system}-{mask}.grib'.format(
                                    dataset=ds, origin=origin,
                                    system=last_system, mask=mask_var
                                )
                                outfile = os.path.join(masks_dir, outfile)

                                downloads.append(( ds, outfile, simplify_req(c) ))

                        else:
                            # FOR ERA5 THE NAMING CONVENTION DIFFERS, THIS NEED TO BE FIXED HERE (example: era5_single_levels_0.25_lsm)
                            
                            # ERA5
                            if 'product_type' in c:
                                for p in c['product_type']:
                                    outfile = '{directory}{dataset}-{product_type}-{mask}.grib'.format(
                                        directory=masks_dir, dataset=ds, product_type=p, mask=mask_var
                                    )
                                    
                                    downloads.append(( ds, outfile, simplify_req(c, prod_type=p) ))
                            
                            # UERRA
                            if 'origin' in c:
                                c['origin'] = c['origin'][0]
                                outfile = '{directory}{dataset}-{origin}-{mask}'.format(
                                    directory=masks_dir, dataset=ds, origin=c['origin'], mask=mask_var
                                )
                                downloads.append(( ds, outfile, simplify_req(c) ))

    # Download unique masks
    # dataset: d[0] , outfile: d[1], info: d[2]
    for d in downloads:
        if not(os.path.exists(d[1])):
            call_download(cred, d[1], d[2], d[0], logdir)
            # saving all downloaded files path
            add_data_for_csv(csv_data, csv_date, outfile)

    create_csv(csv_data, csv_path)

if __name__ == "__main__":
    main()


