#!/usr/bin/env python
#
# Data Quality Checker also known as DQC
#
# Read the README.md for more detailed information
#
#       Barcelona Supercomputing Center - Centro Nacional de Supercomputacion
#       Earth Sciences Department - Computational Earth Sciences
#
#     Authors:
#       Alejandro Alfonso Hernandez (alejandro.alfonso@bsc.es)


import sys
import csv
from pathlib import Path

# Custom packages
from dqc_downloader.utils_download import *

"""
Download MONTHLY seasonal and reanalysis datasets
(depending on the json file : <config_file>)
Try to download the n months (by default n=3, see utils_download.get_dates_to_down) before the current one
depending on the months available in each constraint, except for those that are
already downloaded (according to the corresponding log file)
"""
def main():

    # Params
    if len(sys.argv) != 4:
        print('USAGE: python download_monthly_current_year.py <config_file> <credentials_file> <csv_name>')
        exit()
    else:
        conf_file = sys.argv[1]
        cred_path = sys.argv[2]
        csv_name = sys.argv[3]

    # General configuration
    download_options = 'conf_monthly_updates/download_options_monthly_updates.json'
    download_options = os.path.join(os.path.dirname(os.path.abspath(__file__)), download_options)
    print("download_options: ", download_options)
    with open(download_options) as json_file:
        conf = json.load(json_file)
        outputdir = conf['outdir']
        logdir = conf['logdir']
        csv_dir = conf['csv_dir']
        datasets_info_dir = conf['datasets_info_dir']
        cds_forms_c3s_location = Path(conf['formsdir'])
        
    # URL and key to download
    with open(cred_path) as json_file:
        cred = json.load(json_file)

    if not os.path.exists(csv_dir):
        os.makedirs(csv_dir)
    
    today_date = datetime.now()

    csv_data = defaultdict(list) # dicc that will contain data to be ingested by the csv
    csv_date = today_date.strftime("%Y-%m-%d %H:%M:%S")

    datasets_info_path = os.path.join(datasets_info_dir, conf_file)

    # There are systems of certain originating centers that it has been agreed not to download anymore (Updated on 10/06/2021 according to WP1 QAR release calendar)
    excluded_systems = ['cmcc-35', 'ukmo-15', 'ukmo-600', 'jma-2', 'dwd-21', 'eccc-1', 'eccc-2']

    # Read list of products to download
    with open(datasets_info_path) as json_file:
        data = json.load(json_file)

    # For every dataset
    for dataset in data['dataset']:

        # Get constraints for this dataset
        constraints = get_constraints(dataset, cds_forms_c3s_location, "constraints")

        if data['group'] == 'seasonal':
            origin = data['form']['originating_centre']
        
        variables = find_dataset_variables(constraints)
             
        # For every variable
        for var in variables:
            
            # Load into pandas table
            df = pd.DataFrame()
            for c in constraints:
                df = df.append(c, ignore_index=True)
                
            # Select variable/origin/system
            drops = []
            for ind, row in df.iterrows():
                # Filtering differs by product family
                if data['group'] == 'seasonal':
                    system = sorted(row['system'], key=lambda x: int(x))[-1]
                    # We only download the ones that are not on the excluded list
                    origin_system = origin + '-' + system
                    condition = origin in row['originating_centre'] and system in row['system'] and var in row['variable'] and not origin_system in excluded_systems

                else:
                    condition = var in row['variable']

                # We do not need the rest
                if condition:                      
                    row['variable'] = var
                    if data['group'] == 'seasonal':
                        row['system'] = system
                        row['originating_centre'] = origin
                # Drop other rows
                else:
                    drops.append(ind)

            # Drop at once
            df = df.drop(drops)

            # Group by year/month
            downloads = {}

            for ind, row in df.iterrows():
                years, months = get_row_dates(row)
                for y, y_months in get_dates_to_down(today_date).items():
                    if y in years:
                        for m in y_months:
                            if m in months:
                                # Forced parameters
                                row['month'] = m
                                row['year'] = y
                                
                                key = '{}{}01'.format(y,m)

                                # Changing parameters (unique values)
                                if key in downloads:
                                    downloads[key]['product_type'] = str_to_list(downloads[key]['product_type'])
                                    downloads[key]['product_type'] = list(set(downloads[key]['product_type'] + row['product_type']))
                                    if data['group'] == 'seasonal':
                                        downloads[key]['leadtime_month'] = str_to_list(downloads[key]['leadtime_month'])
                                        downloads[key]['leadtime_month'] = sorted(list(set(downloads[key]['leadtime_month'] + row['leadtime_month'])), key = float)
                                    else:
                                        downloads[key]['time'] = str_to_list(downloads[key]['time'])
                                        downloads[key]['time'] = sorted(list(set(downloads[key]['time'] + row['time'])))

                                    # Pressure levels
                                    if 'pressure_level' in downloads[key]:
                                        downloads[key]['pressure_level'] = str_to_list(downloads[key]['pressure_level'])
                                        downloads[key]['pressure_level'] = sorted(list(set(downloads[key]['pressure_level'] + row['pressure_level'])), key = float)

                                else:
                                    downloads[key] = row.to_dict()

            # Empty products
            if not downloads:
                continue

            # Perform download (GRIB)
            for k, info in downloads.items():
                # Output directory                    
                outdir = os.path.join(outputdir, '{}/{}/{}'.format(data['group'], dataset, info['variable']))                    
                if not(os.path.exists(outdir)):
                    os.makedirs(outdir)
                # Output file
                if data['group'] == 'seasonal':
                    outfile = "{dir}/{origin}-{system}-{date}.{format}".format(
                        dir=outdir, date=k, format='grib', prod=dataset,
                        origin=info['originating_centre'], system=info['system']
                    )
                else:
                    outfile = "{dir}/{var}-{date}.{format}".format(
                        dir=outdir, var=info['variable'], date=k, format='grib', prod=dataset
                    )
                
                # Download perform and save logs
                call_download(cred, outfile, info, dataset, logdir)

                # saving all downloaded files path
                add_data_for_csv(csv_data, csv_date, outfile)

            # NetCDF sample only needed for reanalysis datasets (CDS does not provide seasonal dataset data in NetCDF format)
            if data['group'] == 'reanalysis':
                # Perform ONE download with ONE resolution (NetCDF sample of the converter)
                try:
                    get_netcdf_sample(info, cred, dataset, outfile, logdir)
                    
                    # saving netcdf file path for csv
                    add_data_for_csv(csv_data, csv_date, outfile)
                except:
                    pass # CDS fails to convert sometimes

    csv_path = os.path.join(csv_dir, csv_name)
    create_csv(csv_data, csv_path)


if __name__ == "__main__":
    main()       
        