#!/usr/bin/env python
#
# Data Quality Checker also known as DQC
#
# Read the README.md for more detailed information
#
#       Barcelona Supercomputing Center - Centro Nacional de Supercomputacion
#       Earth Sciences Department - Computational Earth Sciences
#
#     Authors:
#       Joan Sala Calero (joan.sala@bsc.es)
#       Jesus Pena-Izquierdo (jesus.pena@bsc.es) 
#       Ivan Cernicharo Ortiz (ivan.cernicharo@bsc.es)
#       Alejandro Alfonso Hernandez (alejandro.alfonso@bsc.es)
#

import os
import json
import pandas as pd
import cdsapi
import time as t
from datetime import *
from collections import defaultdict
from dateutil.relativedelta import *

from dqc_wrapper.utils.utils_files import get_dates_from_files

"""
Downloader
"""
def call_download(cred, outfile, info, dataset, logdir):
    # Download or skip (do not overwrite)
    if not os.path.exists(outfile):
    #if not exists_in_log(outfile, dataset, logdir):
        print(info)
        print("DOWN = {}/{}".format(dataset, os.path.basename(outfile)))
        u = cred['key'].split(':')[0]
        downloader(dataset, info, outfile, logpath=logdir, key=cred['key'], user=u)
    else:
        print("SKIP = {}/{}".format(dataset, os.path.basename(outfile)))


def downloader(dataset, form, download_file, 
                url = None, key = None,
                logpath = None, user = 'user'):

    #This function uses the cdsapi.Client().retrieve() function to download any specified dataset from the CDS while storing key features of the download
    #in a log file (in csv format) for latter analysis. This log file is stored in the folder specified by logpath. Download features related with a
    # download request for same dataset type and a same user will be added in a same log file. 

    #dataset: same first input as in the cdsapi.Client().retrieve(). It specifies the name of the dataset to be downloaded
    #form: same second input as in the cdsapi.Client().retrieve(). It specifies all the corresponding detail of the download
    #download_file: same third input as in the cdsapi.Client().retrieve(). It specifies the path and file name of the corresponding download file

    #url: url used by the cdsapi.Client(url, key)
    #key: user key used by the cdsapi.Client(url, key)
    
    #logpath: path for the log file of the download features. By the fault it is '/data/logs_cds_download/'. If this folder does not exist it is
    #created.
    #user: the UID of the user to be added to the log file.

    #Definition of logpath in case is not specified
    if logpath is None:
        logpath = '/data/logs_cds_download/'
        if not os.path.exists(logpath):
            os.makedirs(logpath)

    #Date and time of the request
    dt_request = datetime.now()
    time0 = t.time()
    
    #Start the cdsapi download request
    download_error = None
    try:
        c = cdsapi.Client(url = url, key = key)
        c.retrieve(dataset,
                    form,
                    download_file)

        #size of the download file
        file_size = round(os.stat(download_file).st_size/1000./1000.,1)
        
    #if the download fails
    except Exception as e:
        print(e)
        file_size = None
        
        #keep the download error
        download_error = e

    #keep the download time
    time1 = t.time()
    dtime_download = (time1 - time0)/60

    #Generate all the download features
    log_dict = form.copy()
    for id in log_dict.keys():
        log_dict[id] = str(log_dict[id])
    log_dict['dataset_name'] = dataset
    log_dict['file_size'] = file_size
    log_dict['download_error'] = download_error
    log_dict['datetime_request'] = dt_request.strftime('%m%d%y-%H%M%S')
    log_dict['weekday'] = dt_request.weekday()
    log_dict['download_time'] = dtime_download
    log_dict['user'] = user
    log_dict['file_path'] = download_file

    #Generate the log file as a csv file
    logfile = dataset + '.csv'
    logpathfile = logpath + logfile
    log_df = pd.DataFrame(log_dict, index=[0])
    log_df = log_df[sorted(log_df.columns)]
    
    #Add request features to an already log file or create a new one otherwise
    if os.path.exists(logpathfile):
        with open(logpathfile, 'a') as f:
            log_df.to_csv(f, index = False, header = False)
    else:
        log_df.to_csv(logpathfile, index = False)
    


"""
Return full constraints/mapping object
"""
def get_constraints(dataset_name, cds_forms_c3s_location, json_name):

    try:
        json_path = cds_forms_c3s_location.joinpath("cds-forms-c3s", dataset_name, json_name + ".json")
        with open(json_path, "r") as cj:
            json_data = json.load(cj)
    except:
        print('Could not open: {}'.json_path)

    return json_data

"""
Return filtered constraints/mapping object
"""
def get_filtered_constraints(dataset_name, cds_forms_c3s_location, json_name, var, origin, system):

    res = []
    try:
        json_path = cds_forms_c3s_location.joinpath("cds-forms-c3s", dataset_name, json_name + ".json")
        with open(json_path, "r") as cj:
            json_data = json.load(cj)

        # Filter out variable
        for c in json_data:
            if var in c['variable'] and origin in c['originating_centre'] and system in c['system']:
                c['variable'] = var
                res.append(c)
    except:
        print('Could not open: {}'.json_path)

    return res

"""
Get netcdf sample from the C3s converter
"""
def get_netcdf_sample(info, cred, prod, outfile, logdir):

    # Hard-coded of course
    info['format'] = 'netcdf'
    outfile = outfile.replace('.grib', '.nc')

    # List of product types (we need to pick one)
    if 'product_type' in info:
        if isinstance(info['product_type'], type([])):
            info['product_type'] = info['product_type'][0]
        else:
            info['product_type'] = info['product_type']

    call_download(cred, outfile, info, prod, logdir)

"""
Get row (month, year) for monthly datasets
Get row (month, year, day) for daily datasets
"""
def get_row_dates(row, is_daily=False):

    years = row['year']
    months = row['month']

    if isinstance(years, str):
        years = [years]
    if isinstance(months, str):
        months = [months]

    if is_daily:
        days = row['day']
        if isinstance(days, str):
            days = [days]
        return years, months, days

    return years, months

"""
Get diccionary with dates where keys are years and values are lists of months.

- param 'n' is an integer representing the number of months to download before the month of param 'date'
without include it.
- param 'date' is a datetime object.

e.g.
    date = datetime(2021, 2, 1)
    get_dates_to_down(date, n=3)
    
    Output {'2020': ['11', '12'], '2021': ['01']}

"""
def get_dates_to_down(date, n=3):
    
    dates_dict = defaultdict(list)
    
    dates = list(pd.date_range(date-relativedelta(months=+n),date-timedelta(days=1),freq='m'))
    dates = list( map(lambda x: str(x).split(' ')[0], dates) )
    
    for d in dates:
        y = d.split('-')[0]
        m = d.split('-')[1]
        if not(y in dates_dict):
            dates_dict[y] = []
        if not(m in dates_dict[y]):
            dates_dict[y].append(m)
            
    return dict(dates_dict)



"""
Generate a download json object compatible with the cdsapi
"""
def generate_download_obj(constraints, conf, file2dwn):

    # Extract date pattern from file
    dt = get_dates_from_files([file2dwn])
    yr, mn, d = dt[0].strftime('%Y'), dt[0].strftime('%m'), '01'

    # Filter out constraints - Get all product types associated with the constraint + month + variable/origin/system
    selection = {}
    for row in constraints:
        if 'seasonal' in conf.cds_dataset:
            condition = conf.origin in row['originating_centre'] and conf.system in row['system'] \
                        and conf.cds_variable in row['variable'] \
                        and mn in row['month'] and yr in row['year']
        else:
            condition = conf.cds_variable in row['variable'] and mn in row['month'] and yr in row['year']

        # If the condition is met, we need to store all "different" combinations, avoid repetitions
        if condition:
            if selection == {}:
                selection = row
            else:
                for k in row.keys():
                    if type(row[k]) is list:
                        if k in selection:
                            selection[k] = sorted(list(set(row[k]+selection[k])))
                    else:
                        selection[k] = row[k]

    # Fix variable/month/year + system/origin(seasonal)
    selection['variable'] = conf.cds_variable
    selection['month'] = mn
    selection['year'] = yr
    if 'seasonal' in conf.cds_dataset:
        selection['originating_centre'] = conf.origin
        selection['system'] = conf.system

    return selection


"""
Check if variable type is string and if so, return it inside a list
Otherwise, return it as it is
"""
def str_to_list(v):
    result = v
    if isinstance(v, str):
        result = [v]
    return result

"""
Finds all variables available in a dataset
according to its constraints
"""
def find_dataset_variables(constraints):
    variables = []
    for cons in constraints:
        for var in cons['variable']:
            if var not in variables:
                variables.append(var)
    return variables

"""
Ingests diccionary with downloaded date and file path
"""
def add_data_for_csv(csv_data, csv_date, outfile):
    csv_data['downloaded_date'].append(csv_date)
    csv_data['file_path'].append(outfile)


"""
Saves full path of the downloaded files in the csv path

This csv is going to be used to generate ini files
and for the checker (see checker_workflow.py)
"""
def create_csv(csv_data, csv_path):
    open(csv_path, 'w').close() # to clean the content before start appending the data
    df = pd.DataFrame(data=csv_data).drop_duplicates(subset='file_path')
    df.to_csv(csv_path, mode='a', index=False, header=None, sep='\t')


"""
Checks in the log file if the file to be downloaded has not been previously downloaded
returns True if it is already in its corresponding log file
returns False otherwise
"""
def exists_in_log(outfile, dataset, logdir):
    found = False
    for log in os.listdir(logdir):
        if log.endswith('.csv') and dataset in log:
            log_path = os.path.join(logdir, log)
            df = pd.read_csv(log_path, error_bad_lines=False)
            if ( (df['file_path'] == outfile) & (df['download_error'].isna()) ).any():
                found = True
    return found


