#!/usr/bin/env python
#
# Data Quality Checker also known as DQC
#
# Read the README.md for more detailed information
#
#       Barcelona Supercomputing Center - Centro Nacional de Supercomputacion
#       Earth Sciences Department - Computational Earth Sciences
#     
#     Authors:
#       Alejandro Alfonso Hernandez (alejandro.alfonso@bsc.es)

import gc
import sys
import traceback
import json
import os
import glob
import csv
from collections import defaultdict

# Local imports
from dqc_wrapper.utils.utils_forms import get_forms_dates
from dqc_wrapper.utils.utils_files import *
from dqc_wrapper.utils.utils_pdf import *

from dqc_wrapper.test00_file_format import FileFormatCheck
from dqc_wrapper.test01_standard_compliance import StandardComplianceTest
from dqc_wrapper.test02_spatial_completeness import SpatialCompletenessTest
from dqc_wrapper.test03_temporal_completeness import TemporalCompletenessTest
from dqc_wrapper.test04_spatial_consistency import SpatialConsistencyTest
from dqc_wrapper.test05_temporal_consistency import TemporalConsistencyTest
from dqc_wrapper.test06_valid_ranges import ValidRangesTest

def generate_ini(rootd, pattern, variable_e, datatype, variable, units, res, gt, gmn,
                 ts, origin, system, dataset, is_s, dr, tg, com, lsm, lnd, vmin, vmax, cmsname):

    ini='''[general]
input = /shared/ale_cds_downloads/{r}/{ds}/{v}
fpattern = {p}
log_dir = /shared/QAR_dqc/operational/logs  
res_dir = /shared/QAR_dqc/operational/results
forms_dir = /shared/cds-forms-c3s-2021-06-01
masks_dir = /shared/cds_downloads/masks
cms_name = {cms}

[dataset]
variable = {ve}
datatype = {dt}
cds_dataset = {ds}
cds_variable = {v}
cds_units = {un}
complete = {com}
origin = {o}
system = {s}

[file_format]

[standard_compliance_grib]

[standard_compliance_netcdf]

[spatial_completeness]
land_sea_mask = {lsm}
land_variable = {lnd}

[temporal_completeness]
is_seasonal = {is_s}

[spatial_consistency]
grid_interval = {res}
grid_type = {gt}
grid_mapping_name = {gmn}
dual_resolution = {dr}

[temporal_consistency]
time_step = {ts}
time_granularity = {tg}

[valid_ranges]
valid_min = {vmin}
valid_max = {vmax}
'''.format(
    p=pattern, ve=variable_e, dt=datatype, v=variable, is_s=is_s, dr=dr, gt=gt, com=com, gmn=gmn, un=units, cms=cmsname,
    o=origin, s=system, ds=dataset, ts=ts, r=rootd, res=res, tg=tg, lsm=lsm, lnd=lnd, vmin=vmin, vmax=vmax)

    return ini

"""
Get tolerance limits
"""
def get_tolerance_limits(tolerance_file):
    # Read and skip comments
    var_min_max = dict()
    with open(tolerance_file, 'r') as fp:
        rdr = csv.reader(filter(lambda row: row[0]!='#', fp))
        for row in rdr:
            if len(row):
                var_min_max[row[0]] = row[1:3]

    fp.close()
    return var_min_max


"""
Get temporal resolution (if needed) + land_sea_mask
"""
def get_temporal_res(var, temp_resolution):
    var = var.replace('_anomaly', '') # postprocessed
    for v in temp_resolution:
        variable, s, g, l = v.split(',')
        if variable.strip() == var.strip():
            return s, g, l

    return None, None, 'no'

"""
Get grid resolution of corresponding variable (for product_type=reanalysis),
return -1 if var is not found in the file
Note: if there are new variables in the CDS, the file should be updated accordingly
"""
def get_era5_sl_res(var, mask_variables):
    for v in mask_variables:
        variable, res, _, _ = v.split(',')
        if variable.rstrip() == var.rstrip():
            return res
    return -1

"""
Get lsm and lnd of corresponding variable, return None, None if var is not found in the file
Note: if there are new variables in the CDS, the file should be updated accordingly
"""
def get_era5_sl_lsm_lnd(var, mask_variables):
    for v in mask_variables:
        variable, _, lsm, lnd = v.split(',')
        if variable.rstrip() == var.rstrip():
            result = lsm, lnd
    return None, None

"""
Init parameters for a reanalysis dataset
"""
def init_params_reanalysis(origin, d, v, vinfo, mask_variables, tolerance_limits, daily, uerra, land):
    root = 'reanalysis'

    # Get variable info
    ve = vinfo['shortName']
    u = vinfo['c3sunit']

    # Some units we know how to set limits
    if u == '%':
        vmin = 0.0
        vmax = 100.0
    elif u == '(0 - 1)':
        vmin = 0.0
        vmax = 1.0
    elif u.lower() == 'degrees':
        vmin = 0.0
        vmax = 360.0
    else:
        vmin = ''
        vmax = ''

    # Try to get tolerance limits (WMO or based on world records)
    try:
        vmin = float(tolerance_limits[v][0].strip())
    except:
        vmin = ''
    try:
        vmax = float(tolerance_limits[v][1].strip())
    except:
        vmax = ''

    # Model distinction
    try:
        o = origin  # uerra has model distinction
    except:
        o = ''

    # Time resolution granularity (complete or not, split dates or not)
    if daily:
        ts = 3
        com = 'no'
        tg = 'hour'
    else:
        ts = 1
        com = 'yes'
        tg = 'month'
    if uerra:
        if 'mescan' in o:
            ts = 12
        else:
            ts = 6
    if land:
        ts = 1
        if daily:
            tg = 'hour'
            com = 'no'

    # Grid resolution (two resolutions allowed) - 0.25 and 0.5
    gmn = ''
    if uerra:
        res = 11000
        if 'mescan' in o:
            res = 5500
        dr = 'no'
        gt = 'projection'
        gmn = 'lambert_conformal_conic'

        patt = '*.grib'
        if o != '':
            patt = '{}-*.grib'.format(o)
    else:
        model = 'era5'
        res = 0.25
        dr = 'yes'
        gt = 'lonlat'
        if land:
            model = 'era5_land'
            res = 0.1
            dr = 'no'
        patt = '{}-*.grib'.format(v)

    # Product type
    dt = 'an, fc, em'

    # Get lsm and lnd flags and also update grid_res for era5-single-levels if necessary
    if land:
        # era5-land
        lsm = 'yes'
        lnd = 'yes'

    elif uerra:
        # uerra-height, uerra-pressure
        lsm = 'no'
        lnd = 'no'
        if 'soil' in d:
            # uerra-soil
            lsm = 'yes'
            lnd = 'yes'
        if 'single' in d:
            # Only var surface_roughness has lsm (for uerra-single-levels)
            if v == 'surface_roughness':
                lsm = 'yes'
                lnd = 'yes'
    else:
        # era5-pressure-levels
        lsm = 'no'
        lnd = 'no'

        if 'single' in d:
            # Get lsm, lnd and res for era5-single-levels
            has_lsm, is_lnd = get_era5_sl_lsm_lnd(v, mask_variables)
            v_res = get_era5_sl_res(v, mask_variables)
            if v_res != -1:
                lsm = has_lsm
                lnd = is_lnd
                res = v_res
            else:
                # case when variable is not found
                print('ERROR: variable {} not found in dqc_wrapper/QAR_dqc/mask_variables_reanalysis'.format(v))
                # TODO: any error must go to a log file (maybe QAR_dqc_generation.log )


    # Naming ini file and cmsname according to naming convention
    if len(o):
        cmsname = '{}-{}-{}'.format(d.replace('-', '_'), o, v.replace('-', '_'))
    else:
        if uerra:
            # for UERRA height_levels or pressure_levels the origin is always uerra_harmonie
            cmsname = '{}-uerra_harmonie-{}'.format(d.replace('-', '_'), v.replace('-', '_'))

        else: # era_5 / era5_land
            cmsname = '{}-{}-{}'.format(d.replace('-', '_'), model, v.replace('-', '_'))

    qar_name = cmsname + '.ini'



    return qar_name, generate_ini(root, patt, ve, dt, v, u, res, gt, gmn, ts, o, '', d, 'no', dr, tg, com, lsm, lnd, vmin, vmax, cmsname)


"""
Init parameters for a seasonal dataset
"""
def init_params_seasonal(o, s, d, v, vinfo, mask_variables, temp_resolution, tolerance_limits, daily):

    # Get variable info
    ve = vinfo['shortName']
    u = vinfo['c3sunit']

    # Try to get tolerance limits
    try:
        vmin = float(tolerance_limits[v][0].strip())
    except:
        vmin = ''
    try:
        vmax = float(tolerance_limits[v][1].strip())
    except:
        vmax = ''

    # Time resolution + Data type
    com = 'yes'
    lnd = 'no'
    if daily:
        com = 'no'
        is_s = 'daily'
        dt = 'fc, hc, em'
        if 'pressure' in d:
            ts, tg, lsm = '12', 'hour', 'no'
        else:
            ts, tg, lsm = get_temporal_res(v, temp_resolution)
    else:
        ts, tg = '1', 'month'
        is_s = 'monthly'
        dt = 'fcmean, hcmean, em'
        lsm = 'no'
        if not('pressure' in d):
            try:
                _, _, lsm = get_temporal_res(v, temp_resolution)
            except:
                print(d,v)
                exit()

    # Grid resolution
    res = 1
    dr = 'no'
    gt = 'lonlat'

    # Origin/System
    #o = dl['form']['originating_centre']
    #s = dl['form']['system']

    # File pattern
    patt = o + '-' + s + '*.grib'

    # Naming ini file and cmsname according to naming convention
    cmsname = '{}-{}-{}-{}'.format(d.replace('-','_'), o, s, v.replace('-','_'))
    cmsname = cmsname.replace('ukmo', 'uk_met_office')

    qar_name = cmsname + '.ini'

    return qar_name, generate_ini('seasonal', patt, ve, dt, v, u, res, gt, '', ts, o, s, d, is_s, dr, tg, com, lsm, lnd, vmin, vmax, cmsname)


"""
Generate a ini file (datachecker configuration)
for each of dataset/variable combination taking only into account
the downloaded files registered on each csv file from the csv folder
"""
def get_qar_names(csv_path, codes_var, mask_variables, temp_resolution, tolerance_limits):

    qar_f = defaultdict(list)
    qar_ini = {}

    with open(csv_path, "r") as download_files:
        reader = csv.reader(download_files, delimiter="\t")
        for line in reader:        
            output_path = line[-1]
            
            # only if the file exists, we generate its ini file
            if os.path.exists(output_path):

                variable = output_path.rsplit('/')[-2]
                dataset = output_path.rsplit('/')[-3]
                group = output_path.rsplit('/')[-4]
                
                ve = codes_var[dataset]['variables'][variable]

                csv_filename = csv_path.rsplit('/', 1)[1]
            
                # SEASONAL
                if group == 'seasonal':
                    origin = output_path.rsplit('/')[-1].split('-')[0]
                    system = output_path.rsplit('/')[-1].split('-')[1]

                    qar_name, ini = init_params_seasonal(origin, system, dataset, variable, ve, mask_variables, temp_resolution, tolerance_limits, 'daily' in csv_filename)

                # REANALYSIS
                else:
                    # UERRA single-levels or soil-levels
                    if 'uerra' in dataset and ('single-levels' in dataset or 'soil-levels' in dataset):
                        origin = output_path.rsplit('/')[-1].split('-')[0]
                    # ERA5
                    # UERRA pressure-levels or height-levels 
                    else:
                        origin = ''

                    qar_name, ini = init_params_reanalysis(origin, dataset, variable, ve, mask_variables, tolerance_limits, 'daily' in csv_filename, 'uerra' in dataset, 'land' in dataset)

                qar_f[qar_name].append(output_path)
                qar_ini[qar_name] = ini

    return qar_f, qar_ini


"""
- Reads config file as unique commandline operator
- Loops over the different checks/sections
"""
def checker(qar_files, qars_dir):

    for ini_file in qar_files:

        ini_path = os.path.join(qars_dir, ini_file)

        # Read configuration and check mandatory fields in the general section
        c = CheckerConfig(ini_path)
        try:
            c.read_config()
        except Exception as e:
            print('ERROR: {}'.format(e))
            sys.exit()
            
        # Skip if zip file exists
        zipfname = get_zip_fname(c.res_dir, ini_file)

        # For each section/test
        files2zip = []
        files2check = []
        ncfiles2check = []
        
        err = False

        for ff in qar_files[ini_file]:

            if ff.endswith('.grib'):
                files2check.append(ff)
            else:
                ncfiles2check.append(ff) # only a sample

        # For every data check
        for s in c.config.sections():

            c.file_list['original'] = files2check
            c.file_list['nc_converter'] = ncfiles2check

            # Section header
            print_header("Data Quality Checker [check={}]".format(s))

            # Free memory (every test)
            gc.collect()

            # Each section is a different test available
            try:
                test_res = None

                ## TEST 00 - File format (both for original GRIB + converted/experimental NetCDF)
                if s == 'file_format':
                    file_format_ck = FileFormatCheck(c.file_list['original'] + c.file_list['nc_converter'])
                    test_res = file_format_ck.main()
                    del file_format_ck

                ## TEST 010 - Standard compliance (original GRIB files)
                elif s == 'standard_compliance_grib':
                    standard_ck = StandardComplianceTest(c.file_list['original'],
                                                         c.forms_dir, c.cds_dataset, c.cds_variable, suffix='grib')
                    test_res = standard_ck.main()
                    del standard_ck

                ## TEST 011 - Standard compliance (converted/experimental NetCDF)
                elif s == 'standard_compliance_netcdf':
                    standard_ck = StandardComplianceTest(c.file_list['nc_converter'],
                                                         c.forms_dir, c.cds_dataset, c.cds_variable, suffix='netcdf')
                    test_res = standard_ck.main()
                    del standard_ck

                ## TEST 02
                elif s == 'spatial_completeness':
                    spatial_completeness_ck = SpatialCompletenessTest(
                        c.file_list['original'], c.varname, c.datatype, c.cds_dataset, c.system, c.origin, c.masks_dir, c.land_sea_mask, c.land_variable
                    )
                    test_res = spatial_completeness_ck.main()
                    del spatial_completeness_ck

                ## TEST 03
                elif s == 'temporal_completeness':
                    temporal_completeness_ck = TemporalCompletenessTest(
                        c.res_dir, c.config_fname, c.file_list['original'], c.forms_dir,
                        c.cds_dataset, c.cds_variable, c.origin, c.system, c.is_seasonal, c.is_complete, c.time_granularity
                    )
                    test_res = temporal_completeness_ck.main()
                    del temporal_completeness_ck

                ## TEST 04
                elif s == 'spatial_consistency':
                    spatial_consistency_ck = SpatialConsistencyTest(
                        c.file_list['original'], c.grid_type, c.grid_int, c.grid_map, c.dual_res
                    )
                    test_res = spatial_consistency_ck.main()
                    del spatial_consistency_ck

                ## TEST 05
                elif s == 'temporal_consistency':
                    temporal_consistency_ck = TemporalConsistencyTest(
                        c.res_dir, c.config_fname, c.file_list['original'], c.time_step, c.time_granularity, c.is_complete
                    )
                    test_res = temporal_consistency_ck.main()
                    del temporal_consistency_ck

                ## TEST 06
                elif s == 'valid_ranges':
                    valid_ranges_ck = ValidRangesTest(
                        c.res_dir, c.config_fname, c.file_list['original'], c.varname, c.valid_min, c.valid_max, c.datatype,
                        c.forms_dir, c.cds_dataset, c.cds_variable, c.cds_units
                    )
                    test_res = valid_ranges_ck.main()          
                    files2zip.append(test_res['plot'])
                    del valid_ranges_ck

                else:
                    raise Exception('Test {} is not available in the current checker implementation. Please check your config syntax')

            # Exceptions/Errors
            except KeyboardInterrupt:
                print_progress(100, 100)
                err=True
                print('Cancelled, exiting...')
                exit()
            except:            
                print_progress(100, 100)
                err=True
                print('ERROR: Processing section {}'.format(s))
                print('ERROR: {}'.format(traceback.format_exc()))

            # Append result
            finally:
                c.save_config(s, test_res)

        if not(err):

            # Save config and PDF export
            #try:
            # Save TXT result object (ini file, overview of all results)
            res_file = save_result(c.config, c.res_dir, c.config_fname)

            # Save PDF report (result + log)
            if os.path.exists(res_file) and os.path.exists(c.log_file):
                files2zip += save_all_pdf_reports(c, res_file)

            # Zip all results (if completed - 6 tests + 1 html)
            if len(files2zip) >= 7:
                zip_results(zipfname, files2zip)
            #except:
                #print('ERROR: Could not save results. Check res_dir and log_dir permissions')

        logging_reset()

def write_file(outf, string):
    with open(outf, 'w') as f:
        f.write(string)

def main():

    # Params
    if len(sys.argv) != 2:
        print('USAGE: checker_workflow.py <csv_filename>')
        exit()
    else:
        csv_filename = sys.argv[1]

    current_path = os.path.dirname(os.path.abspath(__file__))
    conf_path = os.path.join(current_path, "conf")
    conf_path = os.path.join(conf_path, "checker_workflow_conf.json")
    if not os.path.exists(conf_path):
        print("ERROR: No config file found in {}. Please, check the given path".format(conf_path))
        exit()

    # General configuration
    with open(conf_path) as json_file:
        conf = json.load(json_file)
        datasets = conf['datasets']
        csv_dir = conf['csv_dir']
        masks_path = conf['masks_path']
        temp_res_path = conf['temp_res_path']
        tolerance_path = conf['tolerance_path']
        qars_dir = conf['qars_dir']

    csv_path = os.path.join(csv_dir, csv_filename)

    if not os.path.exists(qars_dir):
        os.makedirs(qars_dir)

    # Get codes
    codes_var = json.load(open(datasets,'r'))

    # Variables that require a land_sea_mask
    mask_variables = []
    with open(masks_path) as f:
        mask_variables = f.read().splitlines()

    # Variables that require a land_sea_mask
    temp_resolution = []
    with open(temp_res_path) as f:
        temp_resolution = f.read().splitlines()

    # Variables that have tolerance limits
    tolerance_limits = get_tolerance_limits(tolerance_path)

    # Get file names
    qar_files, qar_ini = get_qar_names(csv_path, codes_var, mask_variables, temp_resolution, tolerance_limits)
    #print(qar_names)
    
    for q, i in qar_ini.items():

        qar=os.path.join(qars_dir, q)

        # Skip if file exists
        if os.path.exists(qar):
                print('Error, file already exists: {}'.format(qar))
                exit(1)
        else:
            write_file(qar, i)
            print('NEW: {}'.format(qar))

    checker(qar_files, qars_dir)

if __name__ == "__main__":
	main()


